//-----------------------------------------------------------------------------
// File: VertexShader.cpp
//
// Desc: Example code showing how to do vertex shaders in D3D.
//
// Copyright (c) 2000 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
(*----------------------------------------------------------------------------*
 *  NVIDIA Cg Delphi demo                                                     *
 *  This demo is based on "runtime_d3d\VertexShader" sample from              *
 *  NVIDIA Cg toolkit Beta 1                                                  *
 *                                                                            *
 * Portions created by NVidia                           		      *
 *  Copyright (c) 2002 NVIDIA                                                 *
 *                                                                            *
 * Portions created by me                                 		      *
 *  Copyright (c) 2002 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 20-Jun-2002                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit VertexShaderUnit;

interface

uses
  Windows, Messages, d3dres, 
  {$IFDEF DXG_COMPAT}DirectXGraphics{$ELSE}Direct3D8{$ENDIF},
  D3DX8, D3DApp, D3DFile, D3DFont, D3DUtil, DXUtil,
  ///////////////////
  cgD3D;

var
  cg_: TcgDirect3D;

  pVertexProgramContainer : TcgProgramContainer = nil;
  pPixelProgramContainer  : TcgProgramContainer = nil;

  vertex_mat_iter       : PcgBindIter = nil;
  vertex_vA_iter        : PcgBindIter = nil;
  vertex_vD_iter        : PcgBindIter = nil;
  vertex_vSin_iter      : PcgBindIter = nil;
  vertex_vCos_iter      : PcgBindIter = nil;
  vertex_Kd_iter        : PcgBindIter = nil;

//cgBindIter * ppixel_base_color0_iter = 0;
//cgBindIter * ppixel_base_color1_iter = 0;
  ppixel_height_iter    : PcgBindIter = nil;

  tex0_iter     : PcgBindIter = nil;
  tex1_iter     : PcgBindIter = nil;

  pContextContainer: TcgContextContainer = nil;


type
  PDisplayVertex = ^TDisplayVertex;
  TDisplayVertex = packed record
    temp: TD3DXVector2;  // different vb than cg
    pos : TD3DXVector2;
    tex0: TD3DXVector2;
    tex1: TD3DXVector2;
  end;


type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    // Font for drawing text
    m_pFont:            CD3DFont;
    m_pFontSmall:       CD3DFont;

    // Scene
    m_pVB:              IDirect3DVertexBuffer8;
    m_pIB:              IDirect3DIndexBuffer8;
    m_dwNumVertices:    DWord;
    m_dwNumIndices:     DWord;
    //m_dwShader:         DWord;
    m_pNvidiaTexture:   IDirect3DTexture8;
    m_pNvidiaTexture2:  IDirect3DTexture8;

    m_dwSize:           DWord;

    // Transforms
    m_matPosition:      TD3DXMatrix;
    m_matView:          TD3DXMatrix;
    m_matProj:          TD3DXMatrix;

    // Navigation
    m_bKey:             array [0..255] of Byte;
    m_fSpeed:           Single;
    m_fAngularSpeed:    Single;
    m_bShowHelp:        Bool;

    m_vVelocity:        TD3DXVector3;
    m_vAngularVelocity: TD3DXVector3;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function FinalCleanup: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;
    function ConfirmDevice(var pCaps: TD3DCaps8; dwBehavior: DWord; Format: TD3DFormat): HResult; override;
  public
    constructor Create;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM ): LRESULT; override;
  end;


implementation

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create(0);
  m_strWindowTitle   := 'CG D3D-Runtime Example';
  m_bUseDepthBuffer  := True;

  m_pFont            := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFontSmall       := CD3DFont.Create('Arial',  9, D3DFONT_BOLD);
  m_pIB              := nil;
  m_pVB              := nil;
  m_dwSize           := 32;
  m_dwNumIndices     := (m_dwSize - 1) * (m_dwSize - 1) * 6;
  m_dwNumVertices    := m_dwSize * m_dwSize;
//  m_dwShader         := 0;

  FillChar(m_bKey, sizeof(m_bKey), 0);
  m_fSpeed           := 5.0;
  m_fAngularSpeed    := 1.0;
  m_bShowHelp        := False;

  m_vVelocity        := D3DXVector3Zero;
  m_vAngularVelocity := D3DXVector3Zero;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit;
var
  vEye, vAt, vUp: TD3DXVector3;
begin
  cg_:= TcgDirect3D.Create;

  // Setup the view matrix
  vEye := D3DXVector3(2.0, 3.0, 3.0);
  vAt  := D3DXVector3(0.0, 0.0, 0.0);
  vUp  := D3DXVector3(0.0, 1.0, 0.0);
  D3DXMatrixLookAtRH(m_matView, vEye, vAt, vUp);

  // Set the position matrix
  D3DXMatrixInverse(m_matPosition, nil, m_matView);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove;
var
  fSecsPerFrame: Single;
  vT, vR: TD3DXVector3;
  v1, v2: TD3DXVector3; // temporary

  matT, matR: TD3DXMatrix;
  qR: TD3DXQuaternion;

  mat: TD3DXMatrix;
  vA, vD: TD3DXVector4;
  vSin, vCos: TD3DXVector4;

  v: TD3DXVector4;
  height: Single;
begin
  fSecsPerFrame := m_fElapsedTime;

  // Process keyboard input
  vT:= D3DXVector3Zero;
  vR:= D3DXVector3Zero;

  if (m_bKey[VK_LEFT] or m_bKey[VK_NUMPAD1]) <> 0       then vT.x:= vT.x - 1.0; // Slide Left
  if (m_bKey[VK_RIGHT] or m_bKey[VK_NUMPAD3]) <> 0      then vT.x:= vT.x + 1.0; // Slide Right
  if (m_bKey[VK_DOWN]) <> 0                             then vT.y:= vT.y - 1.0; // Slide Down
  if (m_bKey[VK_UP]) <> 0                               then vT.y:= vT.y + 1.0; // Slide Up
  if (m_bKey[Ord('W')]) <> 0                            then vT.z:= vT.z - 2.0; // Move Forward
  if (m_bKey[Ord('S')]) <> 0                            then vT.z:= vT.z + 2.0; // Move Backward
  if (m_bKey[Ord('A')] or m_bKey[VK_NUMPAD8]) <> 0      then vR.x:= vR.x - 1.0; // Pitch Down
  if (m_bKey[Ord('Z')] or m_bKey[VK_NUMPAD2]) <> 0      then vR.x:= vR.x + 1.0; // Pitch Up
  if (m_bKey[Ord('E')] or m_bKey[VK_NUMPAD6]) <> 0      then vR.y:= vR.y - 1.0; // Turn Right
  if (m_bKey[Ord('Q')] or m_bKey[VK_NUMPAD4]) <> 0      then vR.y:= vR.y + 1.0; // Turn Left
  if (m_bKey[VK_NUMPAD9]) <> 0                          then vR.z:= vR.z - 2.0; // Roll CW
  if (m_bKey[VK_NUMPAD7]) <> 0                          then vR.z:= vR.z + 2.0; // Roll CCW

  // m_vVelocity        := m_vVelocity * 0.9 + vT * 0.1;
  D3DXVec3Add(m_vVelocity,
    D3DXVec3Scale(v1, m_vVelocity, 0.9)^,
    D3DXVec3Scale(v2, vT, 0.1)^);
  // m_vAngularVelocity := m_vAngularVelocity * 0.9 + vR * 0.1;
  D3DXVec3Add(m_vAngularVelocity,
    D3DXVec3Scale(v1, m_vAngularVelocity, 0.9)^,
    D3DXVec3Scale(v2, vR, 0.1)^);

  // Update position and view matricies
  // vT := m_vVelocity * fSecsPerFrame * m_fSpeed;
  D3DXVec3Scale(vT, m_vVelocity, fSecsPerFrame * m_fSpeed);
  // vR := m_vAngularVelocity * fSecsPerFrame * m_fAngularSpeed;
  D3DXVec3Scale(vR, m_vAngularVelocity, fSecsPerFrame * m_fAngularSpeed);

  D3DXMatrixTranslation(matT, vT.x, vT.y, vT.z);
  D3DXMatrixMultiply(m_matPosition, matT, m_matPosition);

  D3DXQuaternionRotationYawPitchRoll(qR, vR.y, vR.x, vR.z);
  D3DXMatrixRotationQuaternion(matR, qR);

  D3DXMatrixMultiply( m_matPosition, matR, m_matPosition );
  D3DXMatrixInverse( m_matView, nil, m_matPosition );
  m_pd3dDevice.SetTransform(D3DTS_VIEW, m_matView);

  // Set up the vertex shader constants
  {}
    D3DXMatrixMultiply(mat, m_matView, m_matProj);
    D3DXMatrixTranspose(mat, mat);

    vA:= D3DXVector4(Sin(m_fTime)*15.0, 0.0, 0.5, 1.0);
    vD:= D3DXVector4(D3DX_PI, 1.0/(2.0*D3DX_PI), 2.0*D3DX_PI, 0.05);

    // Taylor series coefficients for sin and cos
    vSin:= D3DXVector4(1.0, -1.0/6.0, 1.0/120.0, -1.0/5040.0);
    vCos:= D3DXVector4(1.0, -1.0/2.0, 1.0/ 24.0, -1.0/ 720.0);

    pVertexProgramContainer.SetShaderConstant(vertex_mat_iter, mat );
    pVertexProgramContainer.SetShaderConstant(vertex_vA_iter,  vA  );
    pVertexProgramContainer.SetShaderConstant(vertex_vD_iter,  vD  );
    pVertexProgramContainer.SetShaderConstant(vertex_vSin_iter,vSin);
    pVertexProgramContainer.SetShaderConstant(vertex_vCos_iter,vCos);
    v:= D3DXVector4(1,0,1,0);
    pVertexProgramContainer.SetShaderConstant(vertex_Kd_iter, v);

//    float base_color0[4] = {1, 1, 1, 1};
//    float base_color1[4] = {1, 0.5, 0.5 ,1};
    //pPixelProgramContainer->SetShaderConstant( ppixel_base_color0_iter, &base_color0);
    //pPixelProgramContainer->SetShaderConstant( ppixel_base_color1_iter, &base_color1);

    height := 1;
    if Assigned(ppixel_height_iter) then
        pPixelProgramContainer.SetShaderConstant(ppixel_height_iter, height);
  {}

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render;
begin
  // Clear the backbuffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
//                       $000000ff, 1.0, 0);
                         $00334C66, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    pPixelProgramContainer.SetTexture(tex0_iter, m_pNvidiaTexture);
    pPixelProgramContainer.SetTexture(tex1_iter, m_pNvidiaTexture2);

    pVertexProgramContainer.SetShaderActive;
    pPixelProgramContainer.SetShaderActive;

    m_pd3dDevice.SetStreamSource(0, m_pVB, SizeOf(TDisplayVertex));
    m_pd3dDevice.SetIndices(m_pIB, 0);
    m_pd3dDevice.DrawIndexedPrimitive(D3DPT_TRIANGLELIST, 0, m_dwNumVertices,
                                      0, m_dwNumIndices div 3);

    //Clootie: Fix to enable correct display in DirectX SDK 8.0 D3DFramework 
    m_pd3dDevice.SetPixelShader(0);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    if m_bShowHelp then
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255),
                            'Keyboard controls:');
      m_pFontSmall.DrawText(20, 60, D3DCOLOR_ARGB(255,255,255,255),
                            'Move'#10'Turn'#10'Pitch'#10'Slide'#10 +
                            'Help'#10'Change device'#10'Exit');
      m_pFontSmall.DrawText(210, 60, D3DCOLOR_ARGB(255,255,255,255),
                            'W,S'#10'E,Q'#10'A,Z'#10'Arrow keys'#10 +
                            'F1'#10'F2'#10'Esc');
    end else
    begin
      m_pFontSmall.DrawText(2, 40, D3DCOLOR_ARGB(255,255,255,255),
                            'Press F1 for help');
    end;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects;
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);
  m_pFontSmall.InitDeviceObjects(m_pd3dDevice);

  Result:= D3DAPPERR_MEDIANOTFOUND;

  // Create texture for the dolphin
  if FAILED(D3DUtil_CreateTexture(m_pd3dDevice, 'gloss_map_nvidia.dds',
                                  m_pNvidiaTexture)) then
  begin
    if FAILED(D3DUtil_CreateTexture(m_pd3dDevice, '..\gloss_map_nvidia.dds',
                                    m_pNvidiaTexture))
    then Exit;
  end;


  if FAILED(D3DUtil_CreateTexture(m_pd3dDevice, 'normal.tga',
                                  m_pNvidiaTexture2)) then
  begin
    if FAILED(D3DUtil_CreateTexture(m_pd3dDevice, '..\normal.tga',
                                    m_pNvidiaTexture2))
    then Exit;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects;
const
  // shows how to use different vertex declaration than cg specifies
  vertex_attributes: array[0..4] of TcgVertexDefinition =
  (
    // number of floats for each element
    (D3DVSDT_type: D3DVSDT_FLOAT2; name: 'temp'; stream: 0),  // dummy
    (D3DVSDT_type: D3DVSDT_FLOAT2; name: 'position'; stream: 0),
    (D3DVSDT_type: D3DVSDT_FLOAT2; name: 'texcoord0'; stream: 0),
    (D3DVSDT_type: D3DVSDT_FLOAT2; name: 'texcoord1'; stream: 0),
    (D3DVSDT_type: -1; name: nil; stream: 0) // CGVertexDefinitionEnd
  );

var
  pIndices: PWord;
  y, x: DWord;
  pVertices: PDisplayVertex;

  fAspectRatio: Single;

  listing: PChar;
  object_code: PChar;
  vert: PChar;
  error_text: PChar;
//  t0, t1: Integer; //Clootie: not used anyway
begin
  vertex_attributes[4]:= CGVertexDefinitionEnd; //Clootie: Init our constant table
  
  m_bReady := True; //Clootie: Why it's where?

  m_pFont.RestoreDeviceObjects;
  m_pFontSmall.RestoreDeviceObjects;

  // Setup render states
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iFALSE);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

  // Create index buffer
  {}
    Result:= m_pd3dDevice.CreateIndexBuffer(m_dwNumIndices*SizeOf(WORD),
                            0, D3DFMT_INDEX16, D3DPOOL_DEFAULT, m_pIB);
    if FAILED(Result) then Exit;

    Result:= m_pIB.Lock(0, m_dwNumIndices*sizeof(WORD), PByte(pIndices), 0);
    if FAILED(Result) then Exit;

    for y:= 1 to m_dwSize - 1 do
    begin
      for x:= 1 to m_dwSize - 1 do
      begin
        pIndices^:= (y-1)*m_dwSize + (x-1); Inc(pIndices);
        pIndices^:= (y-0)*m_dwSize + (x-1); Inc(pIndices);
        pIndices^:= (y-1)*m_dwSize + (x-0); Inc(pIndices);

        pIndices^:= (y-1)*m_dwSize + (x-0); Inc(pIndices);
        pIndices^:= (y-0)*m_dwSize + (x-1); Inc(pIndices);
        pIndices^:= (y-0)*m_dwSize + (x-0); Inc(pIndices);
      end;
    end;

    Result:= m_pIB.Unlock;
    if FAILED(Result) then Exit;
  {}


  // Create vertex buffer
  {}
    Result:= m_pd3dDevice.CreateVertexBuffer(m_dwNumVertices*SizeOf(TDisplayVertex),
                                             0, 0,
                                             D3DPOOL_DEFAULT, m_pVB);
    if FAILED(Result) then Exit;

    Result:= m_pVB.Lock(0, m_dwNumVertices*SizeOf(TDisplayVertex), PByte(pVertices), 0);
    if FAILED(Result) then Exit;

    for y:= 0 to m_dwSize - 1 do
    begin
      for x:= 0 to m_dwSize - 1 do
      begin
        pVertices.temp.x := 0;
        pVertices.temp.y := 0;

        pVertices.tex0 := D3DXVector2((x / (m_dwSize-1)),
                                      (y / (m_dwSize-1)));
        pVertices.tex1 := D3DXVector2((x / (m_dwSize-1)),
                                      (y / (m_dwSize-1)));

        // *pVertices++ = D3DXVECTOR2( ((float)x / (float)(m_dwSize-1) - 0.5f) * D3DX_PI,
        //                             ((float)y / (float)(m_dwSize-1) - 0.5f) * D3DX_PI );
        pVertices.pos := D3DXVector2( (x / (m_dwSize-1) - 0.5) * D3DX_PI,
                                      (y / (m_dwSize-1) - 0.5) * D3DX_PI );
        Inc(pVertices);
      end;
    end;

    Result:= m_pVB.Unlock;
    if FAILED(Result) then Exit;
  {}

  //cg.AddFilePath("C:\\AddYourCGDirectoryHere");
  cg_.AddFilePath('..');

  pContextContainer := cg_.CreateContextContainer(m_pd3dDevice);

  // shows how to use different vertex declaration than cg specifies
  pVertexProgramContainer := pContextContainer.LoadCGProgramFromFile(
      'ripple.cg', 'Ripple Vertex Shader', cgDX8VertexProfile, @vertex_attributes);

  if (pVertexProgramContainer = nil) then
  begin
    // listing in case of an error
    listing := pContextContainer.GetLastListing;

    if (listing = nil) then
      listing := 'Could not find cgc.exe.';

    cg_.NotePad('Failed to Create Vertex Program'#10#10, listing);

    Halt(1);
  end else
  begin
    // vertex shader asm code that is generated
    object_code := pVertexProgramContainer.GetProgramObjectCode;

    // uncomment this to see the vertex code
    //cg_.NotePad('', object_code);


    vert := pVertexProgramContainer.GetVertexDeclaration;

    // this should be how your vertex is defined so it matches the cg
    // uncomment this to see how your cg program is expecting the vertex to be defined
    //cg_.NotePad('cg program expects the vertex be defined like this:'#10, vert);
  end;


  // the are the variables from the ripple.cg program
  vertex_mat_iter := pVertexProgramContainer.GetParameterBindByName('mat');


  vertex_vA_iter := pVertexProgramContainer.GetParameterBindByName('vA');
  vertex_vD_iter := pVertexProgramContainer.GetParameterBindByName('vD');
  vertex_vSin_iter := pVertexProgramContainer.GetParameterBindByName('vSin');
  vertex_vCos_iter := pVertexProgramContainer.GetParameterBindByName('vCos');
  vertex_Kd_iter := pVertexProgramContainer.GetParameterBindByName('Kd');
    
    
  pPixelProgramContainer := pContextContainer.LoadCGProgramFromFile('RipplePixel.cg',
                                                                    'test', cgDX8PixelProfile);

  if (pPixelProgramContainer = nil) then
  begin
    error_text := pContextContainer.GetLastListing;

    //MessageBox(0, "Failed to Create Pixel Program", 0, MB_OK);
    cg_.NotePad('Failed to Create Pixel Program'#10#10#10, error_text);

    Halt(1);
  end else
  begin
    object_code := pPixelProgramContainer.GetProgramObjectCode;

    // uncomment this to see the pixel shader code
    //cg_.NotePad('', object_code);

    tex0_iter := pPixelProgramContainer.GetTextureBindByName('tex0');
    tex1_iter := pPixelProgramContainer.GetTextureBindByName('tex1');
  end;


  //ppixel_base_color0_iter = pPixelProgramContainer->GetParameterBindByName('base_color0');
  //ppixel_base_color1_iter = pPixelProgramContainer->GetParameterBindByName('base_color1');
  ppixel_height_iter := pPixelProgramContainer.GetParameterBindByName('height');

  //Clootie: lines below not really used...
{  t0 := pPixelProgramContainer.GetTexturePosition(tex0_iter);
  t1 := pPixelProgramContainer.GetTexturePosition(tex1_iter); }


  // Set up the projection matrix
  fAspectRatio:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovRH(m_matProj, D3DXToRadian(60.0), fAspectRatio, 0.1, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProj);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFontSmall.InvalidateDeviceObjects;
  m_pIB:= nil;
  m_pVB:= nil;
//  m_pd3dDevice.DeleteVertexShader(m_dwShader);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFontSmall.DeleteDeviceObjects;

  SAFE_RELEASE(m_pNvidiaTexture);
  SAFE_RELEASE(m_pNvidiaTexture2);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup;
begin
  m_pFont.Free;
  m_pFontSmall.Free;

  cg_.Free;
  
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device intialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice;
begin
  if( (dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING)) <> 0 then 
  begin
    if (pCaps.VertexShaderVersion < D3DVS_VERSION(1,0)) then
    begin
      Result:= E_FAIL;
      Exit;
    end;
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc;
begin
  // Handle key presses        
  case uMsg of
    WM_KEYUP:   m_bKey[wParam]:= 0;
    WM_KEYDOWN: m_bKey[wParam]:= 1;
    WM_COMMAND:
    begin
      case LOWORD(wParam) of
        IDM_TOGGLEHELP: m_bShowHelp:= not m_bShowHelp;
      end;
    end;
  end;

  // Pass remaining messages to default handler
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

